# -*- coding: utf-8 -*-
#
# MIT License
# Copyright (c) 2020 Netcloud AG

"""ACI

AciClient for doing Username/Password based RestCalls to the APIC
"""
import logging
import json
import requests
import threading

# The modules are named different in python2/python3...
try:
    from urlparse import urlparse, urlunparse, parse_qsl
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlparse, urlunparse, urlencode, parse_qsl

requests.packages.urllib3.disable_warnings()


class ACI:
    __logger = logging.getLogger(__name__)

    # ==============================================================================
    # constructor
    # ==============================================================================
    def __init__(self, apicIp, apicUser, apicPasword, refresh=False):
        self.__logger.debug('Constructor called')
        self.apicIp = apicIp
        self.apicUser = apicUser
        self.apicPassword = apicPasword

        self.baseUrl = 'https://' + self.apicIp + '/api/'
        self.__logger.debug(f'BaseUrl set to: {self.baseUrl}')

        self.refresh_auto = refresh
        self.refresh_next = None
        self.refresh_thread = None
        self.refresh_offset = 30
        self.session = None
        self.token = None

    def __refresh_session_timer(self, response):
        self.__logger.debug(f'refreshing the token {self.refresh_offset}s before it expires')
        self.refresh_next = int(response.json()['imdata'][0]['aaaLogin']['attributes']['refreshTimeoutSeconds'])
        self.refresh_thread = threading.Timer(self.refresh_next - self.refresh_offset, self.renewCookie)
        self.__logger.debug(f'starting thread to refresh token in {self.refresh_next - self.refresh_offset}s')
        self.refresh_thread.start()

    # ==============================================================================
    # login
    # ==============================================================================
    def login(self) -> bool:
        self.__logger.debug('login called')

        self.session = requests.Session()
        self.__logger.debug('Session Object Created')

        # create credentials structure
        userPass = json.dumps({'aaaUser': {'attributes': {'name': self.apicUser, 'pwd': self.apicPassword}}})

        self.__logger.info(f'Login to apic {self.baseUrl}')
        response = self.session.post(self.baseUrl + 'aaaLogin.json', data=userPass, verify=False, timeout=5)

        # Don't raise an exception for 401
        if response.status_code == 401:
            self.__logger.error(f'Login not possible due to Error: {response.text}')
            self.session = False
            return False

        # Raise a exception for all other 4xx and 5xx status_codes
        response.raise_for_status()

        self.token = response.json()['imdata'][0]['aaaLogin']['attributes']['token']
        self.__logger.debug('Successful get Token from APIC')

        if self.refresh_auto:
            self.__refresh_session_timer(response=response)
        return True

    # ==============================================================================
    # logout
    # ==============================================================================
    def logout(self):
        self.__logger.debug('logout called')
        self.refresh_auto = False
        if self.refresh_thread is not None:
            if self.refresh_thread.is_alive():
                self.__logger.debug('Stoping refresh_auto thread')
                self.refresh_thread.cancel()
        self.postJson(jsonData={'aaaUser': {'attributes': {'name': self.apicUser}}}, url='aaaLogout.json')
        self.__logger.debug('Logout from APIC sucessfull')

    # ==============================================================================
    # renew cookie (aaaRefresh)
    # ==============================================================================
    def renewCookie(self) -> bool:
        self.__logger.debug('Renew Cookie called')
        response = self.session.post(self.baseUrl + 'aaaRefresh.json', verify=False)

        if response.status_code == 200:
            if self.refresh_auto:
                self.__refresh_session_timer(response=response)
            self.token = response.json()['imdata'][0]['aaaLogin']['attributes']['token']
            self.__logger.debug('Successfuly renewed the token')
        else:
            self.token = False
            self.refresh_auto = False
            self.__logger.error(f'Could not renew token. {response.text}')
            response.raise_for_status()
            return False
        return True

    # ==============================================================================
    # getToken
    # ==============================================================================
    def getToken(self) -> str:
        self.__logger.debug('Get Token called')
        return self.token

    # ==============================================================================
    # getJson
    # ==============================================================================
    def getJson(self, uri, subscription=False) -> {}:
        url = self.baseUrl + uri
        self.__logger.debug(f'Get Json called url: {url}')

        if subscription:
            url = '{}?subscription=yes'.format(url)
        response = self.session.get(url, verify=False)

        if response.ok:
            responseJson = response.json()
            self.__logger.debug(f'Successful get Data from APIC: {responseJson}')
            if subscription:
                subscription_id = responseJson['subscriptionId']
                self.__logger.debug(f'Returning Subscription Id: {subscription_id}')
                return subscription_id
            return responseJson['imdata']

        elif response.status_code == 400:
            resp_text = response.json()['imdata'][0]['error']['attributes']['text']
            self.__logger.error(f'Error 400 during get occured: {resp_text}')
            if resp_text == 'Unable to process the query, result dataset is too big':
                # Dataset was too big, we try to grab all the data with pagination
                self.__logger.debug(f'Trying with Pagination, uri: {uri}')
                return self.getJsonPaged(uri)
            return resp_text
        else:
            self.__logger.error(f'Error during get occured: {response.json()}')
            return response.json()

    # ==============================================================================
    # getJson with Pagination
    # ==============================================================================
    def getJsonPaged(self, uri) -> {}:
        url = self.baseUrl + uri
        self.__logger.debug(f'Get Json Pagination called url: {url}')
        parsed_url = urlparse(url)
        parsed_query = parse_qsl(parsed_url.query)

        return_data = []
        page = 0

        while True:
            parsed_query.extend([('page', page), ('page-size', '50000')])
            page += 1
            url_to_call = urlunparse((parsed_url[0], parsed_url[1], parsed_url[2], parsed_url[3],
                                      urlencode(parsed_query), parsed_url[5]))
            response = self.session.get(url_to_call, verify=False)

            if response.ok:
                responseJson = response.json()
                self.__logger.debug(f'Successful get Data from APIC: {responseJson}')
                if responseJson['imdata']:
                    return_data.extend(responseJson['imdata'])
                else:
                    return return_data

            elif response.status_code == 400:
                resp_text = '400: ' + response.json()['imdata'][0]['error']['attributes']['text']
                self.__logger.error(f'Error 400 during get occured: {resp_text}')
                return resp_text

            else:
                self.__logger.error(f'Error during get occured: {response.json()}')
                return False

    # ==============================================================================
    # postJson
    # ==============================================================================
    def postJson(self, jsonData, url='mo.json') -> {}:
        self.__logger.debug(f'Post Json called data: {jsonData}')
        response = self.session.post(self.baseUrl + url, verify=False, data=json.dumps(jsonData, sort_keys=True))
        if response.status_code == 200:
            self.__logger.debug(f'Successful Posted Data to APIC: {response.json()}')
            return response.status_code
        elif response.status_code == 400:
            resp_text = '400: ' + response.json()['imdata'][0]['error']['attributes']['text']
            self.__logger.error(f'Error 400 during get occured: {resp_text}')
            return resp_text
        else:
            self.__logger.error(f'Error during get occured: {response.json()}')
            response.raise_for_status()
            return response.status_code

    # ==============================================================================
    # deleteMo
    # ==============================================================================
    def deleteMo(self, dn) -> int:
        self.__logger.debug(f'Delete Mo called DN: {dn}')
        response = self.session.delete(self.baseUrl + "mo/" + dn + ".json", verify=False)

        # Raise Exception if http Error occurred
        response.raise_for_status()

        return response.status_code

    # ==============================================================================
    # snapshot
    # ==============================================================================
    def snapshot(self, description="snapshot", target_dn="") -> bool:
        self.__logger.debug(f'snapshot called {description}')

        json_payload = [
            {
                "configExportP": {
                    "attributes": {
                        "adminSt": "triggered",
                        "descr": f"by aciClient - {description}",
                        "dn": "uni/fabric/configexp-aciclient",
                        "format": "json",
                        "includeSecureFields": "yes",
                        "maxSnapshotCount": "global-limit",
                        "name": "aciclient",
                        "nameAlias": "",
                        "snapshot": "yes",
                        "targetDn": f"{target_dn}"
                    }
                }
            }
        ]

        response = self.postJson(json_payload)
        if response == 200:
            self.__logger.debug('snapshot created and triggered')
            return True
        else:
            self.__logger.error(f'snapshot creation not succesfull: {response}')
            return False



# -*- coding: utf-8 -*-
#
# MIT License
# Copyright (c) 2020 Netcloud AG

"""AciClient Testing

"""
from requests import RequestException

from aciClient.aci import ACI
import pytest
import time

__BASE_URL = 'testing-apic.ncdev.ch'


def test_login_ok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'asdafa'}}}
    ]})
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    assert aci.login()


def test_token_ok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    token = aci.getToken()
    assert token == 'tokenxyz'


def test_login_401(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'status_code': 401, 'text': 'not allowed'},
                       status_code=401)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    assert not aci.login()


def test_login_404_exception(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'status_code': 404, 'text': 'Not Found'},
                       status_code=404)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    with pytest.raises(RequestException):
        resp = aci.login()

def test_login_refresh_ok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'refreshTimeoutSeconds': '31', 'token':'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaRefresh.json', json={
        'imdata': [
            {
            'aaaLogin': {
                'attributes': {
                    'refreshTimeoutSeconds': '300',
                    'token':'tokenabc'
                    }
                }
            }
        ]})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogout.json', json={'imdata': []}, status_code=200)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown', refresh=True)
    aci.login()
    token = aci.getToken()
    time.sleep(2)
    aci.logout()
    assert token != aci.getToken()

def test_login_refresh_nok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'refreshTimeoutSeconds': '31', 'token':'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaRefresh.json', json={
        'imdata': []}, status_code=403)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown', refresh=True)
    aci.login()
    time.sleep(3)
    token = aci.getToken()
    assert not token

def test_renew_cookie_ok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaRefresh.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]},
                       status_code=200)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    assert aci.renewCookie()


def test_renew_cookie_exception(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaRefresh.json', json={'status_code': 401, 'text': 'Not Allowed'},
                       status_code=401)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    with pytest.raises(RequestException):
        aci.renewCookie()


def test_get_tenant_ok(requests_mock):
    uri = 'mo/uni/tn-common.json'
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.get(f'https://{__BASE_URL}/api/{uri}',
                      json={'imdata': [{'fvTenant': {'attributes':
                                                         {'annotation': '',
                                                          'childAction': '',
                                                          'descr': '',
                                                          'dn': 'uni/tn-common',
                                                          'extMngdBy': '',
                                                          'lcOwn': 'local',
                                                          'modTs': '2020-11-23T15:53:52.014+00:00',
                                                          'monPolDn': 'uni/tn-common/monepg-default',
                                                          'name': 'common',
                                                          'nameAlias': '',
                                                          'ownerKey': '',
                                                          'ownerTag': '',
                                                          'status': '',
                                                          'uid': '0'}}}]
                            })
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.getJson(uri)
    assert 'fvTenant' in resp[0]


def test_get_tenant_ok(requests_mock):
    uri = 'mo/uni/tn-common.json'
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.get(f'https://{__BASE_URL}/api/{uri}',
                      json={'imdata': [{'fvTenant': {'attributes':
                                                         {'annotation': '',
                                                          'childAction': '',
                                                          'descr': '',
                                                          'dn': 'uni/tn-common',
                                                          'extMngdBy': '',
                                                          'lcOwn': 'local',
                                                          'modTs': '2020-11-23T15:53:52.014+00:00',
                                                          'monPolDn': 'uni/tn-common/monepg-default',
                                                          'name': 'common',
                                                          'nameAlias': '',
                                                          'ownerKey': '',
                                                          'ownerTag': '',
                                                          'status': '',
                                                          'uid': '0'}}}]
                            })
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.getJson(uri)
    assert 'fvTenant' in resp[0]


def test_get_tenant_not_found(requests_mock):
    uri = 'mo/uni/tn-commmmmon.json'
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.get(f'https://{__BASE_URL}/api/{uri}',
                      json={'error': 'Not found'}, status_code=404)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.getJson(uri)
    assert 'error' in resp


def test_post_tenant_ok(requests_mock):
    post_data = {'fvTenant': {'attributes':
                                  {'descr': '',
                                   'dn': 'uni/tn-test',
                                   'status': 'modified,created'}}
                 }
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json',
                       json={'imdata': post_data})
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.postJson(post_data)
    assert resp == 200


def test_post_tenant_bad_request(requests_mock):
    post_data = {'fvTenFailFailant': {'attributes':
                                          {'descr': '',
                                           'dn': 'uni/tn-test',
                                           'status': 'modified,created'}}
                 }
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json',
                       json={'imdata': [
                           {'error': {'attributes': {'text': 'tokenxyz'}}}]},
                       status_code=400)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.postJson(post_data)
    assert resp.startswith('400: ')


def test_post_tenant_forbidden_exception(requests_mock):
    post_data = {'fvTenFailFailant': {'attributes':
                                          {'descr': '',
                                           'dn': 'uni/tn-test',
                                           'status': 'modified,created'}}
                 }
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json',
                       json={'imdata': [
                           {'error': {'attributes': {'text': 'tokenxyz'}}}]},
                       status_code=403)
    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    with pytest.raises(RequestException):
        aci.postJson(post_data)


def test_snapshot_ok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json', json={"totalCount": "0", "imdata": []})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})

    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.snapshot(description='unit_test')
    assert resp


def test_snapshot_nok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json',
                       json={"totalCount": "0", "imdata": [{"error": {"attributes": {"text": "Error UnitTest"}}}]},
                       status_code=400)
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})

    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.snapshot(description='unit_test')
    assert not resp

def test_snapshot_tenant_ok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json', json={"totalCount": "0", "imdata": []})
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})

    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.snapshot(description='unit_test', target_dn='/uni/tn-test')
    assert resp


def test_snapshot_tenant_nok(requests_mock):
    requests_mock.post(f'https://{__BASE_URL}/api/mo.json',
                       json={"totalCount": "0", "imdata": [{"error": {"attributes": {"text": "Error UnitTest"}}}]},
                       status_code=400)
    requests_mock.post(f'https://{__BASE_URL}/api/aaaLogin.json', json={'imdata': [
        {'aaaLogin': {'attributes': {'token': 'tokenxyz'}}}
    ]})

    aci = ACI(apicIp=__BASE_URL, apicUser='admin', apicPasword='unkown')
    aci.login()
    resp = aci.snapshot(description='unit_test', target_dn='/uni/tn-test')
    assert not resp



# -*- coding: utf-8 -*-
#
# MIT License
# Copyright (c) 2020 Netcloud AG

"""ACI

AciClient for doing Cert based RestCalls to the APIC

NOTES:
openssl req -new -newkey rsa:2048 -days 36500 -nodes -x509 -keyout apicUser.key -out apicUser.crt
"""
import logging
from OpenSSL import crypto
import base64
import requests
import json


class ACICert:
    __logger = logging.getLogger(__name__)

    # ==============================================================================
    # constructor
    # ==============================================================================
    def __init__(self, apicIp, pkPath, certDn):
        self.__logger.debug(f'Constructor called {apicIp} {pkPath} {certDn}')
        self.apicIp = apicIp
        self.baseUrl = 'https://' + self.apicIp + '/api/'
        self.__logger.debug(f'BaseUrl set to: {self.baseUrl}')
        self.pkey = crypto.load_privatekey(crypto.FILETYPE_PEM, open(pkPath, 'rb').read())
        self.certDn = certDn

    # ==============================================================================
    # packCookies
    # ==============================================================================
    def packCookies(self, content) -> {}:
        sigResult = base64.b64encode(crypto.sign(self.pkey, content, 'sha256')).decode()
        # print('sigResultType', type(sigResult), sigResult)
        return {'APIC-Certificate-Fingerprint': 'fingerprint',
                'APIC-Certificate-Algorithm': 'v1.0',
                'APIC-Certificate-DN': self.certDn,
                'APIC-Request-Signature': sigResult}

    # ==============================================================================
    # getJson
    # ==============================================================================
    def getJson(self, uri) -> {}:
        url = self.baseUrl + uri
        self.__logger.debug(f'Get Json called url: {url}')
        content = 'GET/api/' + uri
        cookies = self.packCookies(content)
        r = requests.get(url, cookies=cookies, verify=False)

        # Raise Exception if http Error occurred
        r.raise_for_status()

        self.__logger.debug(f'Successful get Data from APIC: {r.json()}')
        return r.json()['imdata']

    # ==============================================================================
    # postJson
    # ==============================================================================
    def postJson(self, jsonData):
        url = self.baseUrl + 'mo.json'
        self.__logger.debug(f'Post Json called data: {jsonData}')
        data = json.dumps(jsonData, sort_keys=True)
        content = 'POST/api/mo.json' + json.dumps(jsonData)
        cookies = self.packCookies(content)
        r = requests.post(url, data=data, cookies=cookies, verify=False)

        # Raise Exception if http Error occurred
        r.raise_for_status()

        if r.status_code == 200:
            self.__logger.debug(f'Successful Posted Data to APIC: {r.json()}')
            return r.status_code
        else:
            self.__logger.error(f'Error during get occured: {r.json()}')
            return r.status_code, r.json()['imdata'][0]['error']['attributes']['text']

    # ==============================================================================
    # deleteMo
    # ==============================================================================
    def deleteMo(self, dn):
        self.__logger.debug(f'Delete Mo called DN: {dn}')
        url = self.baseUrl + 'mo/' + dn + '.json'
        content = 'DELETE/api/mo/' + dn + '.json'
        cookies = self.packCookies(content)
        r = requests.delete(url, cookies=cookies, verify=False)

        # Raise Exception if http Error occurred
        r.raise_for_status()

        return r.status_code



from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='aciClient',
      version='1.5',
      description='aci communication helper class',
      url='http://www.netcloud.ch',
      author='mze',
      author_email='nc_dev@netcloud.ch',
      license='MIT',
      packages=['aciClient'],
      install_requires=['requests>=2.26.0 , <3', 'pyOpenSSL>=23.0.0, <24'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      python_requires=">=3.6",
      zip_safe=False)
